package org.bdware.consistency.plugin.raft.algo.storage;

import org.apache.commons.io.FileUtils;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.raft.algo.util.FileUtil;
import org.bdware.sc.bean.ContractRequest;

import java.io.File;
import java.io.IOException;
import java.util.*;


public class LogManager {
    private static final Logger LOGGER = org.apache.logging.log4j.LogManager.getLogger(LogManager.class);

    private final String logDir;
    private final TreeMap<Long, String> segmentStartIndexMap = new TreeMap<>(); // key is start index, value is file name of the segment
    private int maxSegmentFileSize;
    private Segment lastSegment;


    public LogManager(String raftDataDir, int maxSegmentFileSize) {
        this.logDir = raftDataDir + File.separator + "log";
        this.maxSegmentFileSize = maxSegmentFileSize;
        File dir = new File(logDir);
        if (!dir.exists()) {
            dir.mkdirs();
        }
        fillMap(dir);
        if (!segmentStartIndexMap.isEmpty()) {
            lastSegment = Segment.loadSegment(logDir, segmentStartIndexMap.lastEntry().getValue());
            LOGGER.info("SegmentStart not Null!" + this.getLastEntryIndex());
        } else {
            LogEntry first = new LogEntry();
            first.setIndex(0);
            first.setTerm(0);
            first.setCommand(new ContractRequest());
            append(first);
            LOGGER.info("SegmentStart is Null!" + this.getLastEntryIndex());
        }
    }

    public LogEntry getLastEntry() {
        if (lastSegment == null) {
            return null;
        }
        return lastSegment.getLastEntry();
    }

    public long getLastEntryIndex() {
        LogEntry entry = getLastEntry();
        if (entry == null) {
            return -1;
        }
        return entry.getIndex();
    }

    public LogEntry getEntry(long index) {
        if (lastSegment == null || index < 0 || index > lastSegment.getEndIndex()) {
            return null;
        }
        if (index >= lastSegment.getStartIndex() && index <= lastSegment.getEndIndex()) {
            return lastSegment.getEntry(index);
        }
        Segment segment = Segment.loadSegment(logDir, segmentStartIndexMap.floorEntry(index).getValue());
        return segment.getEntry(index);
    }

    public long append(LogEntry entry) {
        long newLastLogIndex = this.getLastEntryIndex();
        newLastLogIndex++;
        byte[] entryBytes = entry.getBytes();
        LOGGER.info("AppendLogEntry: last=" + this.getLastEntryIndex() + " -> toAdd=" + entry.getIndex());

        try {
            if (segmentStartIndexMap.isEmpty() || lastSegment == null ||
                    lastSegment.getFileSize() + entryBytes.length > maxSegmentFileSize) {
                // close file and rename
                if (lastSegment != null) {
                    FileUtil.closeFile(lastSegment.getRandomAccessFile());
                    String newFileName = String.format("%016x-%016x",
                            lastSegment.getStartIndex(), lastSegment.getEndIndex());
                    String newFullFileName = logDir + File.separator + newFileName;
                    File newFile = new File(newFullFileName);
                    String oldFullFileName = logDir + File.separator + lastSegment.getFileName();
                    File oldFile = new File(oldFullFileName);
                    FileUtils.moveFile(oldFile, newFile);
                    segmentStartIndexMap.put(lastSegment.getStartIndex(), newFileName);
                }
                String segmentFileName = "open-" + String.format("%016x", newLastLogIndex);
                lastSegment = new Segment(logDir, segmentFileName);
                segmentStartIndexMap.put(newLastLogIndex, segmentFileName);
            }
            if (entry.getIndex() == 0) {
                entry.setIndex(newLastLogIndex);
            }
            if (entry.getIndex() != newLastLogIndex) {
                LOGGER.warn("received invalid log index: {}", entry.getIndex());
                throw new RuntimeException("received invalid log index: {}" + entry.getIndex());
            }
            lastSegment.append(entry);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return newLastLogIndex;
    }

    public long append(List<LogEntry> entries) {
        long newLastLogIndex = this.getLastEntryIndex();
        for (LogEntry entry : entries) {
            newLastLogIndex = append(entry);
        }
        return newLastLogIndex;
    }

    public List<LogEntry> getEntriesStartAt(long logIndex) {
        Map.Entry<Long, String> floorEntry = segmentStartIndexMap.floorEntry(logIndex);
        if (floorEntry == null) {
            LOGGER.warn("received invalid log index: {}", logIndex);
            throw new RuntimeException("received invalid log index: {}" + logIndex);
        }
        SortedMap<Long, String> tailMap = segmentStartIndexMap.tailMap(floorEntry.getKey());
        List<LogEntry> entries = new ArrayList<>();
        tailMap.forEach((key, value) -> {
            Segment segment = Segment.loadSegment(logDir, value);
            entries.addAll(segment.getEntriesNotLessThan(logIndex));
            segment.close();
        });
        return entries;
    }

    public void deleteEntriesStartAt(long logIndex) {
        LOGGER.info("deleteEntry:" + logIndex);

        Map.Entry<Long, String> floorEntry = segmentStartIndexMap.floorEntry(logIndex - 1);
        long floorKey;
        if (floorEntry == null) {
            floorKey = -1;
            lastSegment = null;
        } else {
            floorKey = floorEntry.getKey();
            if (lastSegment.getStartIndex() != floorKey) {
                lastSegment.close();
                lastSegment = Segment.loadSegment(logDir, floorEntry.getValue());
                lastSegment.setMode("rw");
            }
            lastSegment.deleteEntriesStartAt(logIndex);
            segmentStartIndexMap.put(lastSegment.getStartIndex(), lastSegment.getFileName());
        }
        Map.Entry<Long, String> lastMapEntry = segmentStartIndexMap.lastEntry();
        while (lastMapEntry != null && lastMapEntry.getKey() > floorKey) {
            FileUtil.deleteFile(logDir, lastMapEntry.getValue());
            segmentStartIndexMap.remove(lastMapEntry.getKey());
            lastMapEntry = segmentStartIndexMap.lastEntry();
        }
    }

    public void setMaxSegmentFileSize(int maxSegmentFileSize) {
        this.maxSegmentFileSize = maxSegmentFileSize;
    }

    private void fillMap(File dir) {
        if (dir == null) return;
        File[] files = dir.listFiles();
        if (files == null) {
            LOGGER.warn("get file in dir({}) error", dir.getName());
            return;
        }
        for (File file : files) {
            if (file.isFile()) {
                long startIndexInFile;
                String fileName = file.getName();
                String[] splitArray = fileName.split("-");
                if (splitArray.length != 2) {
                    LOGGER.warn("segment filename[{}] is not valid", fileName);
                    throw new RuntimeException("segment filename[{}] is not valid" + fileName);
                }
                if (splitArray[0].equals("open")) {
                    startIndexInFile = Long.parseLong(splitArray[1], 16);
                } else {
                    startIndexInFile = Long.parseLong(splitArray[0], 16);
                }
                segmentStartIndexMap.put(startIndexInFile, fileName);
            }
        }
    }

}
