package org.bdware.consistency.plugin.raft.algo.util;

import org.apache.commons.io.FileUtils;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.raft.algo.message.IRaftMessage;
import org.bdware.consistency.plugin.raft.algo.storage.LogEntry;
import org.bdware.consistency.plugin.raft.algo.storage.Segment;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.lang.reflect.Method;
import java.nio.channels.FileChannel;
import java.util.zip.CRC32;

public class FileUtil {
    private static final Logger LOGGER = org.apache.logging.log4j.LogManager.getLogger(FileUtil.class);

    public static RandomAccessFile createFile(String dir, String fileName, String mode) {
        try {
            String fullFileName = dir + File.separator + fileName;
            File file = new File(fullFileName);
            file.createNewFile();
            return new RandomAccessFile(file, mode);
        } catch (FileNotFoundException e) {
            LOGGER.warn("file not fount, file={}", fileName);
            throw new RuntimeException("file not found, file=" + fileName);
        } catch (IOException e) {
            LOGGER.warn("create file error, msg={}", e.getMessage());
            throw new RuntimeException("create file error, msg={}" + e.getMessage());
        }
    }

    public static RandomAccessFile openFile(String dir, String fileName, String mode) {
        try {
            String fullFileName = dir + File.separator + fileName;
            File file = new File(fullFileName);
            return new RandomAccessFile(file, mode);
        } catch (FileNotFoundException e) {
            LOGGER.warn("file not fount, file={}", fileName);
            throw new RuntimeException("file not found, file=" + fileName);
        }
    }


    public static void closeFile(RandomAccessFile randomAccessFile) {
        try {
            if (randomAccessFile != null) {
                randomAccessFile.close();
            }
        } catch (IOException e) {
            LOGGER.warn("close file error, msg={}", e.getMessage());
            throw new RuntimeException("close file error, msg={}" + e.getMessage());
        }
    }

    public static void deleteFile(String dir, String fileName) {
        String fullFileName = dir + File.separator + fileName;
        File file = new File(fullFileName);
        try {
            FileUtils.forceDelete(file);
        } catch (IOException e) {
            LOGGER.warn("delete file error, msg={}", e.getMessage());
            throw new RuntimeException("delete file error, msg={}" + e.getMessage());
        }
    }

    public static void moveFile(String oldFullFileName, String newFullFileName) {
        File oldFile = new File(oldFullFileName);
        File newFile = new File(newFullFileName);
        try {
            FileUtils.moveFile(oldFile, newFile);
        } catch (IOException e) {
            LOGGER.warn("move file error, msg={}", e.getMessage());
            throw new RuntimeException("move file error, msg={}" + e.getMessage());
        }
    }

    public static void truncate(RandomAccessFile randomAccessFile, long size) {
        FileChannel fileChannel = randomAccessFile.getChannel();
        try {
            fileChannel.truncate(size);
        } catch (IOException e) {
            LOGGER.warn("truncate file error, msg={}", e.getMessage());
            throw new RuntimeException("truncate file error, msg={}" + e.getMessage());
        } finally {
            try {
                fileChannel.close();
            } catch (IOException e) {
                LOGGER.warn("close fileChannel error, msg={}", e.getMessage());
                throw new RuntimeException("close fileChannel error, msg={}" + e.getMessage());
            }
        }
    }

    public static LogEntry readLogEntryFromFile(RandomAccessFile raf) {
        try {
            long crc32FromFile = raf.readLong();
            int dataLen = raf.readInt();
            int hasReadLen = (Long.SIZE + Integer.SIZE) / Byte.SIZE;
            if (raf.length() - hasReadLen < dataLen) {
                LOGGER.warn("file remainLength < dataLen");
                return null;
            }
            byte[] data = new byte[dataLen];
            int readLen = raf.read(data);
            if (readLen != dataLen) {
                LOGGER.warn("readLen != dataLen");
                return null;
            }
            long crc32FromData = getCRC32(data);
            if (crc32FromFile != crc32FromData) {
                LOGGER.warn("crc32 check failed");
                return null;
            }
            return LogEntry.parse(data);
        } catch (Exception ex) {
            LOGGER.warn("readLogEntryFromFile meet exception, {}", ex.getMessage());
            return null;
        }
    }

    public static void writeLogEntryToFile(RandomAccessFile raf, LogEntry entry) {
        byte[] bytes = entry.getBytes();
        long crc32 = getCRC32(bytes);
        try {
            raf.writeLong(crc32);
            raf.writeInt(bytes.length);
            raf.write(bytes);
        } catch (IOException ex) {
            LOGGER.warn("write LogEntry to file error, msg={}", ex.getMessage());
            throw new RuntimeException("write LogEntry to file error");
        }
    }

    public static long getCRC32(byte[] data) {
        CRC32 crc32 = new CRC32();
        crc32.update(data);
        return crc32.getValue();
    }

}
